from collections.abc import Callable
from typing import Any, TypeVar

import openai
from pydantic import BaseModel
from pydantic import ConfigDict as _ConfigDict
from pydantic import with_config as _with_config


class ConfigDict(_ConfigDict, total=False):
    openai_strict: bool


T = TypeVar("T", bound=type | Callable[..., Any])


# Relax type constraints to allow as function decorator
def with_config(config: ConfigDict) -> Callable[[T], T]:
    return _with_config(config)  # type: ignore[return-value]


def get_pydantic_config(obj: Any) -> ConfigDict | None:
    return getattr(obj, "__pydantic_config__", None)


# TODO: Use directly from pydantic when available
# https://github.com/pydantic/pydantic/issues/5213
def json_schema(model: type[BaseModel]) -> dict[str, Any]:
    """Generates a JSON schema for a pydantic model.

    If `openai_strict` is set to `True` in the model's config, a schema compatible with
    OpenAI APIs strict mode is generated.
    """
    if model.model_config.get("openai_strict", False):
        tool_param = openai.pydantic_function_tool(model)
        return tool_param["function"].get("parameters", {})
    model_schema = model.model_json_schema().copy()
    model_schema.pop("title", None)
    model_schema.pop("description", None)
    return model_schema



import os
from typing import Annotated, Any
from unittest.mock import ANY

import openai
import pytest
from inline_snapshot import snapshot
from openai.types.chat import ChatCompletionMessageParam
from pydantic import AfterValidator, BaseModel

from magentic._pydantic import ConfigDict, with_config
from magentic._streamed_response import AsyncStreamedResponse, StreamedResponse
from magentic.chat_model.base import ToolSchemaParseError
from magentic.chat_model.message import (
    AssistantMessage,
    FunctionResultMessage,
    ImageBytes,
    ImageUrl,
    Message,
    SystemMessage,
    Usage,
    UserMessage,
    _RawMessage,
)
from magentic.chat_model.openai_chat_model import (
    OpenaiChatModel,
    async_message_to_openai_message,
    message_to_openai_message,
)
from magentic.function_call import (
    AsyncParallelFunctionCall,
    FunctionCall,
    ParallelFunctionCall,
)
from magentic.streaming import AsyncStreamedStr, StreamedStr, async_iter


def plus(a: int, b: int) -> int:
    return a + b


message_to_openai_message_test_cases = [
    (
        _RawMessage({"role": "user", "content": "Hello"}),
        {"role": "user", "content": "Hello"},
    ),
    (SystemMessage("Hello"), {"role": "system", "content": "Hello"}),
    (UserMessage("Hello"), {"role": "user", "content": "Hello"}),
    (
        UserMessage([ImageUrl("https://example.com/image.jpg")]),
        {
            "role": "user",
            "content": [
                {
                    "type": "image_url",
                    "image_url": {"url": "https://example.com/image.jpg"},
                }
            ],
        },
    ),
    (AssistantMessage("Hello"), {"role": "assistant", "content": "Hello"}),
    (
        AssistantMessage(42),
        {
            "role": "assistant",
            "tool_calls": [
                {
                    "id": ANY,
                    "type": "function",
                    "function": {"name": "return_int", "arguments": '{"value":42}'},
                }
            ],
        },
    ),
    (
        AssistantMessage(FunctionCall(plus, 1, 2)),
        {
            "role": "assistant",
            "tool_calls": [
                {
                    "id": ANY,
                    "type": "function",
                    "function": {"name": "plus", "arguments": '{"a":1,"b":2}'},
                }
            ],
        },
    ),
    (
        AssistantMessage(
            ParallelFunctionCall([FunctionCall(plus, 1, 2), FunctionCall(plus, 3, 4)])
        ),
        {
            "role": "assistant",
            "tool_calls": [
                {
                    "id": ANY,
                    "type": "function",
                    "function": {"name": "plus", "arguments": '{"a":1,"b":2}'},
                },
                {
                    "id": ANY,
                    "type": "function",
                    "function": {"name": "plus", "arguments": '{"a":3,"b":4}'},
                },
            ],
        },
    ),
    (
        AssistantMessage(
            StreamedResponse([StreamedStr(["Hello"]), FunctionCall(plus, 1, 2)])
        ),
        {
            "role": "assistant",
            "content": "Hello",
            "tool_calls": [
                {
                    "id": ANY,
                    "type": "function",
                    "function": {"name": "plus", "arguments": '{"a":1,"b":2}'},
                },
            ],
        },
    ),
    (
        FunctionResultMessage(3, FunctionCall(plus, 1, 2)),
        {
            "role": "tool",
            "tool_call_id": ANY,
            "content": '{"value":3}',
        },
    ),
]


@pytest.mark.parametrize(
    ("message", "expected_openai_message"), message_to_openai_message_test_cases
)
def test_message_to_openai_message(message, expected_openai_message):
    assert message_to_openai_message(message) == expected_openai_message


async_message_to_openai_message_test_cases = [
    *message_to_openai_message_test_cases,
    (
        AssistantMessage(
            AsyncParallelFunctionCall(
                async_iter([FunctionCall(plus, 1, 2), FunctionCall(plus, 3, 4)])
            )
        ),
        {
            "role": "assistant",
            "tool_calls": [
                {
                    "id": ANY,
                    "type": "function",
                    "function": {"name": "plus", "arguments": '{"a":1,"b":2}'},
                },
                {
                    "id": ANY,
                    "type": "function",
                    "function": {"name": "plus", "arguments": '{"a":3,"b":4}'},
                },
            ],
        },
    ),
    (
        AssistantMessage(
            AsyncStreamedResponse(
                async_iter(
                    [
                        AsyncStreamedStr(async_iter(["Hello", " World"])),
                        FunctionCall(plus, 1, 2),
                    ]
                )
            )
        ),
        {
            "role": "assistant",
            "content": "Hello World",
            "tool_calls": [
                {
                    "id": ANY,
                    "type": "function",
                    "function": {"name": "plus", "arguments": '{"a":1,"b":2}'},
                },
            ],
        },
    ),
]


@pytest.mark.parametrize(
    ("message", "expected_openai_message"), async_message_to_openai_message_test_cases
)
async def test_async_message_to_openai_message(message, expected_openai_message):
    assert await async_message_to_openai_message(message) == expected_openai_message


def test_message_to_openai_message_user_image_message_bytes_jpg(image_bytes_jpg):
    image_message = UserMessage([ImageBytes(image_bytes_jpg)])
    assert message_to_openai_message(image_message) == snapshot(
        {
            "role": "user",
            "content": [
                {
                    "type": "image_url",
                    "image_url": {
                        "url": "data:image/jpeg;base64,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"
                    },
                }
            ],
        }
    )


def test_message_to_openai_message_user_image_message_bytes_png(image_bytes_png):
    image_message = UserMessage([ImageBytes(image_bytes_png)])
    assert message_to_openai_message(image_message) == snapshot(
        {
            "role": "user",
            "content": [
                {
                    "type": "image_url",
                    "image_url": {
                        "url": "data:image/png;base64, "
                    },
                }
            ],
        }
    )


def test_message_to_openai_message_raises():
    class CustomMessage(Message[str]):
        def __init__(self, content: str, **data: Any):
            super().__init__(content, **data)

        def format(self, **kwargs):
            del kwargs
            return CustomMessage(self.content)

    with pytest.raises(NotImplementedError):
        message_to_openai_message(CustomMessage("Hello"))

    @message_to_openai_message.register
    def _(message: CustomMessage) -> ChatCompletionMessageParam:
        return {"role": "user", "content": message.content}

    assert message_to_openai_message(CustomMessage("Hello")) == {
        "role": "user",
        "content": "Hello",
    }


@pytest.mark.openai
def test_openai_chat_model_api_key(monkeypatch):
    openai_api_key = os.environ["OPENAI_API_KEY"]
    monkeypatch.delenv("OPENAI_API_KEY")

    with pytest.raises(openai.OpenAIError):
        chat_model = OpenaiChatModel("gpt-4o")

    chat_model = OpenaiChatModel("gpt-4o", api_key=openai_api_key)
    message = chat_model.complete(messages=[UserMessage("Say hello!")])
    assert isinstance(message.content, str)


@pytest.mark.openai
def test_openai_chat_model_complete_base_url():
    chat_model = OpenaiChatModel("gpt-4o", base_url="https://api.openai.com/v1")
    message = chat_model.complete(messages=[UserMessage("Say hello!")])
    assert isinstance(message.content, str)


@pytest.mark.openai
def test_openai_chat_model_complete_seed():
    chat_model = OpenaiChatModel("gpt-4o", seed=42)
    message1 = chat_model.complete(messages=[UserMessage("Say hello!")])
    message2 = chat_model.complete(messages=[UserMessage("Say hello!")])
    assert message1.content == message2.content


@pytest.mark.openai
def test_openai_chat_model_complete_streamed_response():
    def get_weather(location: str) -> None:
        """Get the weather for a location."""

    chat_model = OpenaiChatModel("gpt-4o")
    message = chat_model.complete(
        messages=[UserMessage("Tell me your favorite city. Then get its weather.")],
        functions=[get_weather],
        output_types=[StreamedResponse],
    )
    assert isinstance(message.content, StreamedResponse)
    response_items = list(message.content)
    assert len(response_items) == 2
    streamed_str, function_call = response_items
    assert isinstance(streamed_str, StreamedStr)
    assert len(streamed_str.to_string()) > 1  # Check StreamedStr was cached
    assert isinstance(function_call, FunctionCall)
    assert function_call() is None  # Check FunctionCall is successfully called


@pytest.mark.openai
def test_openai_chat_model_complete_pydantic_model_openai_strict():
    class CapitalCity(BaseModel):
        model_config = ConfigDict(openai_strict=True)
        capital: str
        country: str

    chat_model = OpenaiChatModel("gpt-4o")
    message = chat_model.complete(
        messages=[UserMessage("What is the capital of Ireland?")],
        output_types=[CapitalCity],
    )
    assert isinstance(message.content, CapitalCity)


@pytest.mark.openai
def test_openai_chat_model_complete_function_call_openai_strict():
    @with_config(ConfigDict(openai_strict=True))
    def plus(a: int, b: int) -> int:
        """Sum two numbers."""
        return a + b

    chat_model = OpenaiChatModel("gpt-4o")
    message = chat_model.complete(
        messages=[UserMessage("Use the tool to sum 1 and 2")],
        functions=[plus],
        output_types=[FunctionCall[int]],
    )
    assert isinstance(message.content, FunctionCall)


@pytest.mark.openai
def test_openai_chat_model_complete_usage():
    chat_model = OpenaiChatModel("gpt-4o")
    message = chat_model.complete(
        messages=[UserMessage("Say hello!")], output_types=[StreamedStr]
    )
    str(message.content)  # Finish the stream
    assert isinstance(message.usage, Usage)
    assert message.usage.input_tokens > 0
    assert message.usage.output_tokens > 0


@pytest.mark.openai
def test_openai_chat_model_complete_usage_structured_output():
    chat_model = OpenaiChatModel("gpt-4o")
    message = chat_model.complete(
        messages=[UserMessage("Count to 5")], output_types=[list[int]]
    )
    assert isinstance(message.usage, Usage)
    assert message.usage.input_tokens > 0
    assert message.usage.output_tokens > 0


@pytest.mark.openai
def test_openai_chat_model_complete_no_structured_output_error():
    chat_model = OpenaiChatModel("gpt-4o")
    # Should not raise StructuredOutputError because forced to make tool call
    message: Message[int | bool] = chat_model.complete(
        messages=[
            UserMessage("Tell me a short joke. Return a string, not a tool call.")
        ],
        output_types=[int, bool],
    )
    assert isinstance(message.content, int | bool)


@pytest.mark.openai
def test_openai_chat_model_complete_raises_tool_schema_parse_error():
    def raise_error(v):
        raise ValueError(v)

    class Test(BaseModel):
        value: Annotated[int, AfterValidator(raise_error)]

    chat_model = OpenaiChatModel("gpt-4o")
    with pytest.raises(ToolSchemaParseError):
        chat_model.complete(
            messages=[UserMessage("Return a test value of 42.")],
            output_types=[Test],
        )


@pytest.mark.openai
def test_openai_chat_model_complete_image_bytes(image_bytes_jpg):
    chat_model = OpenaiChatModel("gpt-4o")
    message = chat_model.complete(
        messages=[
            UserMessage(
                ("Describe this image in one word.", ImageBytes(image_bytes_jpg))
            )
        ]
    )
    assert isinstance(message.content, str)


@pytest.mark.openai
async def test_openai_chat_model_acomplete_async_streamed_response():
    def get_weather(location: str) -> None:
        """Get the weather for a location."""

    chat_model = OpenaiChatModel("gpt-4o")
    message = await chat_model.acomplete(
        messages=[UserMessage("Tell me your favorite city. Then get its weather.")],
        functions=[get_weather],
        output_types=[AsyncStreamedResponse],
    )
    assert isinstance(message.content, AsyncStreamedResponse)
    response_items = [x async for x in message.content]
    assert len(response_items) == 2
    streamed_str, function_call = response_items
    assert isinstance(streamed_str, AsyncStreamedStr)
    assert len(await streamed_str.to_string()) > 1  # Check AsyncStreamedStr was cached
    assert isinstance(function_call, FunctionCall)
    assert function_call() is None  # Check FunctionCall is successfully called


@pytest.mark.openai
async def test_openai_chat_model_acomplete_usage():
    chat_model = OpenaiChatModel("gpt-4o")
    message = await chat_model.acomplete(
        messages=[UserMessage("Say hello!")], output_types=[AsyncStreamedStr]
    )
    await message.content.to_string()  # Finish the stream
    assert isinstance(message.usage, Usage)
    assert message.usage.input_tokens > 0
    assert message.usage.output_tokens > 0


@pytest.mark.openai
async def test_openai_chat_model_acomplete_usage_structured_output():
    chat_model = OpenaiChatModel("gpt-4o")
    message = await chat_model.acomplete(
        messages=[UserMessage("Count to 5")], output_types=[list[int]]
    )
    assert isinstance(message.usage, Usage)
    assert message.usage.input_tokens > 0
    assert message.usage.output_tokens > 0


@pytest.mark.openai
async def test_openai_chat_model_acomplete_raises_tool_schema_parse_error():
    def raise_error(v):
        raise ValueError(v)

    class Test(BaseModel):
        value: Annotated[int, AfterValidator(raise_error)]

    chat_model = OpenaiChatModel("gpt-4o")
    with pytest.raises(ToolSchemaParseError):
        await chat_model.acomplete(
            messages=[UserMessage("Return a test value of 42.")],
            output_types=[Test],
        )


def test_openai_chat_model_azure_omits_stream_options(monkeypatch):
    monkeypatch.setenv("AZURE_OPENAI_API_KEY", "test")
    monkeypatch.setenv("AZURE_OPENAI_ENDPOINT", "test")
    monkeypatch.setenv("OPENAI_API_VERSION", "test")
    chat_model = OpenaiChatModel("gpt-4o", api_type="azure")
    assert chat_model._get_stream_options() == openai.NOT_GIVEN



"""Functions for parsing and checking return types."""

from collections.abc import Iterable

from magentic._streamed_response import AsyncStreamedResponse, StreamedResponse
from magentic.function_call import AsyncParallelFunctionCall, ParallelFunctionCall
from magentic.streaming import AsyncStreamedStr, StreamedStr
from magentic.typing import is_any_origin_subclass


def contains_string_type(types: Iterable[type]) -> bool:
    return is_any_origin_subclass(
        types,
        (str, StreamedStr, AsyncStreamedStr, StreamedResponse, AsyncStreamedResponse),
    )


def contains_parallel_function_call_type(types: Iterable[type]) -> bool:
    return is_any_origin_subclass(
        types,
        (
            ParallelFunctionCall,
            AsyncParallelFunctionCall,
            StreamedResponse,
            AsyncStreamedResponse,
        ),
    )



import pytest

from magentic.chat_model.message import UserMessage
from magentic.chat_model.openai_chat_model import OpenaiChatModel


@pytest.mark.parametrize(
    ("prompt", "output_types", "expected_output_type"),
    [
        ("Say hello!", [str], str),
        ("Return True.", [bool], bool),
        ("Return [1, 2, 3, 4, 5]", [list[int]], list),
    ],
)
@pytest.mark.openai_ollama
def test_openai_chat_model_complete_ollama(prompt, output_types, expected_output_type):
    chat_model = OpenaiChatModel("llama3.1", base_url="http://localhost:11434/v1/")
    message = chat_model.complete(
        messages=[UserMessage(prompt)], output_types=output_types
    )
    assert isinstance(message.content, expected_output_type)


@pytest.mark.parametrize(
    ("prompt", "output_types", "expected_output_type"),
    [
        ("Say hello!", [str], str),
        ("Return True.", [bool], bool),
        ("Return [1, 2, 3, 4, 5]", [list[int]], list),
    ],
)
@pytest.mark.openai_ollama
async def test_openai_chat_model_acomplete_ollama(
    prompt, output_types, expected_output_type
):
    chat_model = OpenaiChatModel("llama3.1", base_url="http://localhost:11434/v1/")
    message = await chat_model.acomplete(
        messages=[UserMessage(prompt)], output_types=output_types
    )
    assert isinstance(message.content, expected_output_type)



from typing_extensions import assert_type

from magentic.formatting import BulletedDict, BulletedList, NumberedDict, NumberedList


def test_bulleted_list():
    items = BulletedList(["foo", "bar", "baz"])
    assert_type(items, BulletedList[str])
    assert f"{items}" == "- foo\n- bar\n- baz"


def test_numbered_list():
    items = NumberedList(["foo", "bar", "baz"])
    assert_type(items, NumberedList[str])
    assert f"{items}" == "1. foo\n2. bar\n3. baz"


def test_bulleted_dict():
    items = BulletedDict({"foo": 1, "bar": 2, "baz": 3})
    assert_type(items, BulletedDict[str, int])
    assert f"{items}" == "- foo: 1\n- bar: 2\n- baz: 3"


def test_numbered_dict():
    items = NumberedDict({"foo": 1, "bar": 2, "baz": 3})
    assert_type(items, NumberedDict[str, int])
    assert f"{items}" == "1. foo: 1\n2. bar: 2\n3. baz: 3"



from collections.abc import AsyncIterator, Callable, Iterable, Iterator, Sequence
from enum import Enum
from functools import singledispatch
from typing import Any, Generic, Literal, TypeVar, cast, overload

import openai
from openai.lib.streaming.chat import ChatCompletionStreamState
from openai.types.chat import (
    ChatCompletionChunk,
    ChatCompletionContentPartParam,
    ChatCompletionMessageParam,
    ChatCompletionMessageToolCallParam,
    ChatCompletionNamedToolChoiceParam,
    ChatCompletionStreamOptionsParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolParam,
    ChatCompletionUserMessageParam,
)

from magentic._parsing import contains_parallel_function_call_type, contains_string_type
from magentic._streamed_response import AsyncStreamedResponse, StreamedResponse
from magentic.chat_model.base import ChatModel, aparse_stream, parse_stream
from magentic.chat_model.function_schema import (
    BaseFunctionSchema,
    FunctionCallFunctionSchema,
    function_schema_for_type,
    get_async_function_schemas,
    get_function_schemas,
)
from magentic.chat_model.message import (
    AssistantMessage,
    ImageBytes,
    ImageUrl,
    Message,
    SystemMessage,
    ToolResultMessage,
    Usage,
    UserMessage,
    _RawMessage,
)
from magentic.chat_model.stream import (
    AsyncOutputStream,
    FunctionCallChunk,
    OutputStream,
    StreamParser,
    StreamState,
)
from magentic.function_call import (
    AsyncParallelFunctionCall,
    FunctionCall,
    ParallelFunctionCall,
    _create_unique_id,
)
from magentic.streaming import AsyncStreamedStr, StreamedStr
from magentic.vision import UserImageMessage


class OpenaiMessageRole(Enum):
    ASSISTANT = "assistant"
    SYSTEM = "system"
    TOOL = "tool"
    USER = "user"


@singledispatch
def message_to_openai_message(message: Message[Any]) -> ChatCompletionMessageParam:
    """Convert a Message to an OpenAI message."""
    # TODO: Add instructions for registering new Message type to this error message
    raise NotImplementedError(type(message))


@singledispatch
async def async_message_to_openai_message(
    message: Message[Any],
) -> ChatCompletionMessageParam:
    """Async version of `message_to_openai_message`."""
    return message_to_openai_message(message)


@message_to_openai_message.register(_RawMessage)
def _(message: _RawMessage[Any]) -> ChatCompletionMessageParam:
    assert isinstance(message.content, dict)
    assert "role" in message.content
    assert "content" in message.content
    return cast(ChatCompletionMessageParam, message.content)


@message_to_openai_message.register
def _(message: SystemMessage) -> ChatCompletionMessageParam:
    return {"role": OpenaiMessageRole.SYSTEM.value, "content": message.content}


@message_to_openai_message.register(UserMessage)
def _(message: UserMessage[Any]) -> ChatCompletionUserMessageParam:
    if isinstance(message.content, str):
        return {"role": OpenaiMessageRole.USER.value, "content": message.content}
    if isinstance(message.content, Iterable):
        content: list[ChatCompletionContentPartParam] = []
        for block in message.content:
            if isinstance(block, str):
                content.append({"type": "text", "text": block})
            elif isinstance(block, ImageBytes):
                content.append(
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:{block.mime_type};base64,{block.as_base64()}"
                        },
                    }
                )
            elif isinstance(block, ImageUrl):
                content.append({"type": "image_url", "image_url": {"url": block.root}})
            else:
                msg = f"Invalid block type: {type(block)}"
                raise TypeError(msg)
        return {"role": OpenaiMessageRole.USER.value, "content": content}
    msg = f"Invalid content type: {type(message.content)}"
    raise TypeError(msg)


@message_to_openai_message.register(UserImageMessage)
def _(message: UserImageMessage[Any]) -> ChatCompletionUserMessageParam:
    if isinstance(message.content, bytes):
        image_bytes = ImageBytes(message.content)
        url = f"data:{image_bytes.mime_type};base64,{image_bytes.as_base64()}"
    elif isinstance(message.content, str):
        url = message.content
    else:
        msg = f"Invalid content type: {type(message.content)}"
        raise TypeError(msg)

    return {
        "role": OpenaiMessageRole.USER.value,
        "content": [{"type": "image_url", "image_url": {"url": url, "detail": "auto"}}],
    }


def _function_call_to_tool_call_block(
    function_call: FunctionCall[Any],
) -> ChatCompletionMessageToolCallParam:
    function_schema = FunctionCallFunctionSchema(function_call.function)
    return {
        "id": function_call._unique_id,
        "type": "function",
        "function": {
            "name": function_schema.name,
            "arguments": function_schema.serialize_args(function_call),
        },
    }


@message_to_openai_message.register(AssistantMessage)
def _(message: AssistantMessage[Any]) -> ChatCompletionMessageParam:
    if isinstance(message.content, str):
        return {"role": OpenaiMessageRole.ASSISTANT.value, "content": message.content}

    if isinstance(message.content, FunctionCall):
        return {
            "role": OpenaiMessageRole.ASSISTANT.value,
            "tool_calls": [_function_call_to_tool_call_block(message.content)],
        }

    if isinstance(message.content, ParallelFunctionCall):
        return {
            "role": OpenaiMessageRole.ASSISTANT.value,
            "tool_calls": [
                _function_call_to_tool_call_block(function_call)
                for function_call in message.content
            ],
        }

    if isinstance(message.content, StreamedResponse):
        content: list[str] = []
        function_calls: list[FunctionCall[Any]] = []
        for item in message.content:
            if isinstance(item, StreamedStr):
                content.append(item.to_string())
            elif isinstance(item, FunctionCall):
                function_calls.append(item)
        return {
            "role": OpenaiMessageRole.ASSISTANT.value,
            "content": " ".join(content),
            "tool_calls": [
                _function_call_to_tool_call_block(function_call)
                for function_call in function_calls
            ],
        }

    function_schema = function_schema_for_type(type(message.content))
    return {
        "role": OpenaiMessageRole.ASSISTANT.value,
        "tool_calls": [
            {
                # Can be random because no result will be inserted back into the chat
                "id": _create_unique_id(),
                "type": "function",
                "function": {
                    "name": function_schema.name,
                    "arguments": function_schema.serialize_args(message.content),
                },
            }
        ],
    }


@async_message_to_openai_message.register(AssistantMessage)
async def _(message: AssistantMessage[Any]) -> ChatCompletionMessageParam:
    if isinstance(message.content, AsyncParallelFunctionCall):
        return {
            "role": OpenaiMessageRole.ASSISTANT.value,
            "tool_calls": [
                _function_call_to_tool_call_block(function_call)
                async for function_call in message.content
            ],
        }

    if isinstance(message.content, AsyncStreamedResponse):
        content: list[str] = []
        function_calls: list[FunctionCall[Any]] = []
        async for item in message.content:
            if isinstance(item, AsyncStreamedStr):
                content.append(await item.to_string())
            elif isinstance(item, FunctionCall):
                function_calls.append(item)
        return {
            "role": OpenaiMessageRole.ASSISTANT.value,
            "content": " ".join(content),
            "tool_calls": [
                _function_call_to_tool_call_block(function_call)
                for function_call in function_calls
            ],
        }
    return message_to_openai_message(message)


@message_to_openai_message.register(ToolResultMessage)
def _(message: ToolResultMessage[Any]) -> ChatCompletionMessageParam:
    if isinstance(message.content, str):
        content = message.content
    else:
        function_schema = function_schema_for_type(type(message.content))
        content = function_schema.serialize_args(message.content)
    return {
        "role": OpenaiMessageRole.TOOL.value,
        "tool_call_id": message.tool_call_id,
        "content": content,
    }


# TODO: Use ToolResultMessage to solve this at magentic level
def _add_missing_tool_calls_responses(
    messages: list[ChatCompletionMessageParam],
) -> list[ChatCompletionMessageParam]:
    """Add null responses for tool calls without a response.

    This is required by OpenAI's API.
    "An assistant message with 'tool_calls' must be followed by tool messages responding to each 'tool_call_id'."
    """
    new_messages: list[ChatCompletionMessageParam] = []
    current_tool_call_responses: set[str] = set()
    for message in reversed(messages):
        if tool_call_id := message.get("tool_call_id"):
            current_tool_call_responses.add(tool_call_id)  # type: ignore[arg-type]
        elif tool_calls := message.get("tool_calls"):
            for tool_call in tool_calls:  # type: ignore[attr-defined]
                if tool_call["id"] not in current_tool_call_responses:
                    new_messages.append(
                        {
                            "role": OpenaiMessageRole.TOOL.value,
                            "tool_call_id": tool_call["id"],
                            "content": "null",
                        }
                    )
                    current_tool_call_responses.add(tool_call["id"])
            current_tool_call_responses = set()

        new_messages.append(message)

    return list(reversed(new_messages))


T = TypeVar("T")
BaseFunctionSchemaT = TypeVar("BaseFunctionSchemaT", bound=BaseFunctionSchema[Any])


class BaseFunctionToolSchema(Generic[BaseFunctionSchemaT]):
    def __init__(self, function_schema: BaseFunctionSchemaT):
        self._function_schema = function_schema

    def as_tool_choice(self) -> ChatCompletionNamedToolChoiceParam:
        return {"type": "function", "function": {"name": self._function_schema.name}}

    def to_dict(self) -> ChatCompletionToolParam:
        return {"type": "function", "function": self._function_schema.dict()}


class OpenaiStreamParser(StreamParser[ChatCompletionChunk]):
    def is_content(self, item: ChatCompletionChunk) -> bool:
        return bool(item.choices and item.choices[0].delta.content)

    def get_content(self, item: ChatCompletionChunk) -> str | None:
        if item.choices and item.choices[0].delta.content:
            return item.choices[0].delta.content
        return None

    def is_tool_call(self, item: ChatCompletionChunk) -> bool:
        return bool(item.choices and item.choices[0].delta.tool_calls)

    def iter_tool_calls(self, item: ChatCompletionChunk) -> Iterator[FunctionCallChunk]:
        if item.choices and item.choices[0].delta.tool_calls:
            for tool_call in item.choices[0].delta.tool_calls:
                if tool_call.function:
                    yield FunctionCallChunk(
                        id=tool_call.id,
                        name=tool_call.function.name,
                        args=tool_call.function.arguments,
                    )


class OpenaiStreamState(StreamState[ChatCompletionChunk]):
    """Tracks the state of the OpenAI model output stream.

    - message snapshot
    - usage
    - stop reason
    """

    def __init__(self) -> None:
        self._chat_completion_stream_state = ChatCompletionStreamState(
            input_tools=openai.NOT_GIVEN,
            response_format=openai.NOT_GIVEN,
        )
        self.usage_ref: list[Usage] = []

        # Keep track of tool call index to add this to Mistral tool calls
        self._current_tool_call_index: int = -1
        self._seen_tool_call_ids: set[str] = set()

    def update(self, item: ChatCompletionChunk) -> None:
        # Add tool call index for Mistral tool calls to make compatible with OpenAI
        # TODO: Remove this fix when MistralChatModel switched to mistral python package
        if item.choices:
            for tool_call_chunk in item.choices[0].delta.tool_calls or []:
                if (
                    tool_call_chunk.id is not None
                    and tool_call_chunk.id not in self._seen_tool_call_ids
                ):
                    self._current_tool_call_index += 1
                    self._seen_tool_call_ids.add(tool_call_chunk.id)
                tool_call_chunk.index = self._current_tool_call_index
        self._chat_completion_stream_state.handle_chunk(item)
        if item.usage:
            assert not self.usage_ref
            self.usage_ref.append(
                Usage(
                    input_tokens=item.usage.prompt_tokens,
                    output_tokens=item.usage.completion_tokens,
                )
            )

    @property
    def current_message_snapshot(self) -> Message[Any]:
        snapshot = self._chat_completion_stream_state.current_completion_snapshot
        message = snapshot.choices[0].message
        # TODO: Possible to return AssistantMessage here?
        return _RawMessage(message.model_dump())


def _if_given(value: T | None) -> T | openai.NotGiven:
    return value if value is not None else openai.NOT_GIVEN


R = TypeVar("R")


class OpenaiChatModel(ChatModel):
    """An LLM chat model that uses the `openai` python package."""

    def __init__(
        self,
        model: str,
        *,
        api_key: str | None = None,
        api_type: Literal["openai", "azure"] = "openai",
        base_url: str | None = None,
        max_tokens: int | None = None,
        seed: int | None = None,
        temperature: float | None = None,
    ):
        self._model = model
        self._api_key = api_key
        self._api_type = api_type
        self._base_url = base_url
        self._max_tokens = max_tokens
        self._seed = seed
        self._temperature = temperature

        match api_type:
            case "openai":
                self._client = openai.OpenAI(api_key=api_key, base_url=base_url)
                self._async_client = openai.AsyncOpenAI(
                    api_key=api_key, base_url=base_url
                )
            case "azure":
                self._client = openai.AzureOpenAI(
                    api_key=api_key,
                    base_url=base_url,  # type: ignore[arg-type]
                )
                self._async_client = openai.AsyncAzureOpenAI(
                    api_key=api_key,
                    base_url=base_url,  # type: ignore[arg-type]
                )

    @property
    def model(self) -> str:
        return self._model

    @property
    def api_key(self) -> str | None:
        return self._api_key

    @property
    def api_type(self) -> Literal["openai", "azure"]:
        return self._api_type

    @property
    def base_url(self) -> str | None:
        return self._base_url

    @property
    def max_tokens(self) -> int | None:
        return self._max_tokens

    @property
    def seed(self) -> int | None:
        return self._seed

    @property
    def temperature(self) -> float | None:
        return self._temperature

    def _get_stream_options(self) -> ChatCompletionStreamOptionsParam | openai.NotGiven:
        if self.api_type == "azure":
            return openai.NOT_GIVEN
        return {"include_usage": True}

    @staticmethod
    def _get_tool_choice(
        *,
        tool_schemas: Sequence[BaseFunctionToolSchema[Any]],
        output_types: Iterable[type],
    ) -> ChatCompletionToolChoiceOptionParam | openai.NotGiven:
        """Create the tool choice argument."""
        if contains_string_type(output_types):
            return openai.NOT_GIVEN
        if len(tool_schemas) == 1:
            return tool_schemas[0].as_tool_choice()
        return "required"

    def _get_parallel_tool_calls(
        self, *, tools_specified: bool, output_types: Iterable[type]
    ) -> bool | openai.NotGiven:
        if not tools_specified:  # Enforced by OpenAI API
            return openai.NOT_GIVEN
        if self.api_type == "azure":
            return openai.NOT_GIVEN
        if contains_parallel_function_call_type(output_types):
            return openai.NOT_GIVEN
        return False

    @overload
    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: None = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[str]: ...

    @overload
    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: Iterable[type[R]] = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[R]: ...

    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[R]] | None = None,
        *,
        stop: list[str] | None = None,
        # TODO: Add type hint for function call ?
    ) -> AssistantMessage[str] | AssistantMessage[R]:
        """Request an LLM message."""
        if output_types is None:
            output_types = cast(Iterable[type[R]], [] if functions else [str])

        function_schemas = get_function_schemas(functions, output_types)
        tool_schemas = [BaseFunctionToolSchema(schema) for schema in function_schemas]

        response: Iterator[ChatCompletionChunk] = self._client.chat.completions.create(
            model=self.model,
            messages=_add_missing_tool_calls_responses(
                [message_to_openai_message(m) for m in messages]
            ),
            max_tokens=_if_given(self.max_tokens),
            seed=_if_given(self.seed),
            stop=_if_given(stop),
            stream=True,
            stream_options=self._get_stream_options(),
            temperature=_if_given(self.temperature),
            tools=[schema.to_dict() for schema in tool_schemas] or openai.NOT_GIVEN,
            tool_choice=self._get_tool_choice(
                tool_schemas=tool_schemas, output_types=output_types
            ),
            parallel_tool_calls=self._get_parallel_tool_calls(
                tools_specified=bool(tool_schemas), output_types=output_types
            ),
        )
        stream = OutputStream(
            response,
            function_schemas=function_schemas,
            parser=OpenaiStreamParser(),
            state=OpenaiStreamState(),
        )
        return AssistantMessage._with_usage(
            parse_stream(stream, output_types), usage_ref=stream.usage_ref
        )

    @overload
    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: None = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[str]: ...

    @overload
    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: Iterable[type[R]] = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[R]: ...

    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[R]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> AssistantMessage[R] | AssistantMessage[str]:
        """Async version of `complete`."""
        if output_types is None:
            output_types = [] if functions else cast(list[type[R]], [str])

        function_schemas = get_async_function_schemas(functions, output_types)
        tool_schemas = [BaseFunctionToolSchema(schema) for schema in function_schemas]

        response: AsyncIterator[
            ChatCompletionChunk
        ] = await self._async_client.chat.completions.create(
            model=self.model,
            messages=_add_missing_tool_calls_responses(
                [await async_message_to_openai_message(m) for m in messages]
            ),
            max_tokens=_if_given(self.max_tokens),
            seed=_if_given(self.seed),
            stop=_if_given(stop),
            stream=True,
            stream_options=self._get_stream_options(),
            temperature=_if_given(self.temperature),
            tools=[schema.to_dict() for schema in tool_schemas] or openai.NOT_GIVEN,
            tool_choice=self._get_tool_choice(
                tool_schemas=tool_schemas, output_types=output_types
            ),
            parallel_tool_calls=self._get_parallel_tool_calls(
                tools_specified=bool(tool_schemas), output_types=output_types
            ),
        )
        stream = AsyncOutputStream(
            response,
            function_schemas=function_schemas,
            parser=OpenaiStreamParser(),
            state=OpenaiStreamState(),
        )
        return AssistantMessage._with_usage(
            await aparse_stream(stream, output_types), usage_ref=stream.usage_ref
        )



from collections.abc import Callable, Iterable, Sequence
from typing import Any, Literal, TypeVar, cast, overload

import openai
from openai.lib.streaming.chat import ChatCompletionStreamState
from openai.types.chat import ChatCompletionNamedToolChoiceParam

from magentic._parsing import contains_string_type
from magentic.chat_model.base import ChatModel, aparse_stream, parse_stream
from magentic.chat_model.function_schema import (
    get_async_function_schemas,
    get_function_schemas,
)
from magentic.chat_model.message import AssistantMessage, Message, Usage, _RawMessage
from magentic.chat_model.openai_chat_model import (
    BaseFunctionToolSchema,
    message_to_openai_message,
)
from magentic.chat_model.stream import (
    AsyncOutputStream,
    FunctionCallChunk,
    OutputStream,
    StreamParser,
    StreamState,
)

try:
    import litellm
    from litellm.litellm_core_utils.streaming_handler import (  # type: ignore[attr-defined]
        StreamingChoices,
    )
    from litellm.types.utils import ModelResponse
except ImportError as error:
    msg = "To use LitellmChatModel you must install the `litellm` package using `pip install 'magentic[litellm]'`."
    raise ImportError(msg) from error


class LitellmStreamParser(StreamParser[ModelResponse]):
    def is_content(self, item: ModelResponse) -> bool:
        assert isinstance(item.choices[0], StreamingChoices)
        return bool(item.choices[0].delta.content)

    def get_content(self, item: ModelResponse) -> str | None:
        assert isinstance(item.choices[0], StreamingChoices)
        assert isinstance(item.choices[0].delta.content, str | None)
        return item.choices[0].delta.content

    def is_tool_call(self, item: ModelResponse) -> bool:
        assert isinstance(item.choices[0], StreamingChoices)
        return bool(item.choices[0].delta.tool_calls)

    def iter_tool_calls(self, item: ModelResponse) -> Iterable[FunctionCallChunk]:
        assert isinstance(item.choices[0], StreamingChoices)
        if item.choices and item.choices[0].delta.tool_calls:
            for tool_call in item.choices[0].delta.tool_calls:
                if tool_call.function:
                    yield FunctionCallChunk(
                        id=tool_call.id,
                        name=tool_call.function.name,
                        args=tool_call.function.arguments,
                    )


class LitellmStreamState(StreamState[ModelResponse]):
    def __init__(self) -> None:
        self._chat_completion_stream_state = ChatCompletionStreamState(
            input_tools=openai.NOT_GIVEN,
            response_format=openai.NOT_GIVEN,
        )
        self.usage_ref: list[Usage] = []

    def update(self, item: ModelResponse) -> None:
        # Patch attributes required inside ChatCompletionStreamState.handle_chunk
        if not hasattr(item, "usage"):
            # litellm requires usage is not None for its total usage calculation
            item.usage = litellm.Usage()  # type: ignore[attr-defined]
        if not hasattr(item, "refusal"):
            assert isinstance(item.choices[0], StreamingChoices)
            item.choices[0].delta.refusal = None  # type: ignore[attr-defined]
        self._chat_completion_stream_state.handle_chunk(item)  # type: ignore[arg-type]
        usage = cast(litellm.Usage, item.usage)  # type: ignore[attr-defined,name-defined]
        # Ignore usages with 0 tokens
        if usage and usage.prompt_tokens and usage.completion_tokens:
            assert not self.usage_ref
            self.usage_ref.append(
                Usage(
                    input_tokens=usage.prompt_tokens,
                    output_tokens=usage.completion_tokens,
                )
            )

    @property
    def current_message_snapshot(self) -> Message[Any]:
        snapshot = self._chat_completion_stream_state.current_completion_snapshot
        message = snapshot.choices[0].message
        # Fix incorrectly concatenated role
        message.role = "assistant"
        # TODO: Possible to return AssistantMessage here?
        return _RawMessage(message.model_dump())


R = TypeVar("R")


class LitellmChatModel(ChatModel):
    """An LLM chat model that uses the `litellm` python package."""

    def __init__(
        self,
        model: str,
        *,
        api_base: str | None = None,
        max_tokens: int | None = None,
        metadata: dict[str, Any] | None = None,
        temperature: float | None = None,
        custom_llm_provider: str | None = None,
    ):
        self._model = model
        self._api_base = api_base
        self._max_tokens = max_tokens
        self._metadata = metadata
        self._temperature = temperature
        self._custom_llm_provider = custom_llm_provider

    @property
    def model(self) -> str:
        return self._model

    @property
    def api_base(self) -> str | None:
        return self._api_base

    @property
    def max_tokens(self) -> int | None:
        return self._max_tokens

    @property
    def metadata(self) -> dict[str, Any] | None:
        return self._metadata

    @property
    def temperature(self) -> float | None:
        return self._temperature

    @property
    def custom_llm_provider(self) -> str | None:
        return self._custom_llm_provider

    @staticmethod
    def _get_tool_choice(
        *,
        tool_schemas: Sequence[BaseFunctionToolSchema[Any]],
        output_types: Iterable[type[R]],
    ) -> ChatCompletionNamedToolChoiceParam | Literal["required"] | None:
        """Create the tool choice argument."""
        if contains_string_type(output_types):
            return None
        if len(tool_schemas) == 1:
            return tool_schemas[0].as_tool_choice()
        return "required"

    @overload
    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: None = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[str]: ...

    @overload
    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: Iterable[type[R]] = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[R]: ...

    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[R]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> AssistantMessage[str] | AssistantMessage[R]:
        """Request an LLM message."""
        if output_types is None:
            output_types = cast(Iterable[type[R]], [] if functions else [str])

        function_schemas = get_function_schemas(functions, output_types)
        tool_schemas = [BaseFunctionToolSchema(schema) for schema in function_schemas]

        response = litellm.completion(
            model=self.model,
            messages=[message_to_openai_message(m) for m in messages],
            api_base=self.api_base,
            custom_llm_provider=self.custom_llm_provider,
            max_tokens=self.max_tokens,
            metadata=self.metadata,
            stop=stop,
            stream=True,
            # TODO: Add usage for LitellmChatModel
            temperature=self.temperature,
            tools=[schema.to_dict() for schema in tool_schemas] or None,
            tool_choice=self._get_tool_choice(
                tool_schemas=tool_schemas, output_types=output_types
            ),  # type: ignore[arg-type,unused-ignore]
        )
        assert not isinstance(response, ModelResponse)
        stream = OutputStream(
            stream=response,
            function_schemas=function_schemas,
            parser=LitellmStreamParser(),
            state=LitellmStreamState(),
        )
        return AssistantMessage(parse_stream(stream, output_types))

    @overload
    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: None = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[str]: ...

    @overload
    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: Iterable[type[R]] = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[R]: ...

    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[R]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> AssistantMessage[str] | AssistantMessage[R]:
        """Async version of `complete`."""
        if output_types is None:
            output_types = cast(Iterable[type[R]], [] if functions else [str])

        function_schemas = get_async_function_schemas(functions, output_types)
        tool_schemas = [BaseFunctionToolSchema(schema) for schema in function_schemas]

        response = await litellm.acompletion(
            model=self.model,
            messages=[message_to_openai_message(m) for m in messages],
            api_base=self.api_base,
            custom_llm_provider=self.custom_llm_provider,
            max_tokens=self.max_tokens,
            metadata=self.metadata,
            stop=stop,
            stream=True,
            # TODO: Add usage for LitellmChatModel
            temperature=self.temperature,
            tools=[schema.to_dict() for schema in tool_schemas] or None,
            tool_choice=self._get_tool_choice(
                tool_schemas=tool_schemas, output_types=output_types
            ),  # type: ignore[arg-type,unused-ignore]
        )
        assert not isinstance(response, ModelResponse)
        stream = AsyncOutputStream(
            stream=response,
            function_schemas=function_schemas,
            parser=LitellmStreamParser(),
            state=LitellmStreamState(),
        )
        return AssistantMessage(await aparse_stream(stream, output_types))



import pytest

from magentic.backend import get_chat_model
from magentic.chat_model.anthropic_chat_model import AnthropicChatModel
from magentic.chat_model.litellm_chat_model import LitellmChatModel
from magentic.chat_model.message import AssistantMessage, UserMessage
from magentic.chat_model.mistral_chat_model import MistralChatModel
from magentic.chat_model.openai_chat_model import OpenaiChatModel


def test_backend_anthropic_chat_model(monkeypatch):
    monkeypatch.setenv("MAGENTIC_BACKEND", "anthropic")
    monkeypatch.setenv("MAGENTIC_ANTHROPIC_MODEL", "claude-3-haiku-20240307")
    monkeypatch.setenv("MAGENTIC_ANTHROPIC_API_KEY", "sk-1234567890")
    monkeypatch.setenv("MAGENTIC_ANTHROPIC_BASE_URL", "http://localhost:8080")
    monkeypatch.setenv("MAGENTIC_ANTHROPIC_MAX_TOKENS", "10")
    monkeypatch.setenv("MAGENTIC_ANTHROPIC_TEMPERATURE", "2")
    chat_model = get_chat_model()
    assert isinstance(chat_model, AnthropicChatModel)
    assert chat_model.model == "claude-3-haiku-20240307"
    assert chat_model.api_key == "sk-1234567890"
    assert chat_model.base_url == "http://localhost:8080"
    assert chat_model.max_tokens == 10
    assert chat_model.temperature == 2


def test_backend_mistral_chat_model(monkeypatch):
    monkeypatch.setenv("MAGENTIC_BACKEND", "mistral")
    monkeypatch.setenv("MAGENTIC_MISTRAL_MODEL", "mistral-large-latest")
    monkeypatch.setenv("MAGENTIC_MISTRAL_API_KEY", "sk-1234567890")
    monkeypatch.setenv("MAGENTIC_MISTRAL_BASE_URL", "http://localhost:8080")
    monkeypatch.setenv("MAGENTIC_MISTRAL_MAX_TOKENS", "1024")
    monkeypatch.setenv("MAGENTIC_MISTRAL_SEED", "42")
    monkeypatch.setenv("MAGENTIC_MISTRAL_TEMPERATURE", "2")
    chat_model = get_chat_model()
    assert isinstance(chat_model, MistralChatModel)
    assert chat_model.model == "mistral-large-latest"
    assert chat_model.api_key == "sk-1234567890"
    assert chat_model.base_url == "http://localhost:8080"
    assert chat_model.max_tokens == 1024
    assert chat_model.seed == 42
    assert chat_model.temperature == 2


def test_backend_openai_chat_model(monkeypatch):
    monkeypatch.setenv("MAGENTIC_BACKEND", "openai")
    monkeypatch.setenv("MAGENTIC_OPENAI_MODEL", "gpt-4")
    monkeypatch.setenv("MAGENTIC_OPENAI_API_KEY", "sk-1234567890")
    monkeypatch.setenv("MAGENTIC_OPENAI_API_TYPE", "openai")
    monkeypatch.setenv("MAGENTIC_OPENAI_BASE_URL", "http://localhost:8080")
    monkeypatch.setenv("MAGENTIC_OPENAI_MAX_TOKENS", "1024")
    monkeypatch.setenv("MAGENTIC_OPENAI_SEED", "42")
    monkeypatch.setenv("MAGENTIC_OPENAI_TEMPERATURE", "2")
    chat_model = get_chat_model()
    assert isinstance(chat_model, OpenaiChatModel)
    assert chat_model.model == "gpt-4"
    assert chat_model.api_key == "sk-1234567890"
    assert chat_model.api_type == "openai"
    assert chat_model.base_url == "http://localhost:8080"
    assert chat_model.max_tokens == 1024
    assert chat_model.seed == 42
    assert chat_model.temperature == 2


def test_backend_litellm_chat_model(monkeypatch):
    monkeypatch.setenv("MAGENTIC_BACKEND", "litellm")
    monkeypatch.setenv("MAGENTIC_LITELLM_API_BASE", "http://localhost:11434")
    monkeypatch.setenv("MAGENTIC_LITELLM_MODEL", "claude-2")
    monkeypatch.setenv("MAGENTIC_LITELLM_MAX_TOKENS", "1024")
    monkeypatch.setenv("MAGENTIC_LITELLM_TEMPERATURE", "2")
    chat_model = get_chat_model()
    assert isinstance(chat_model, LitellmChatModel)
    assert chat_model.api_base == "http://localhost:11434"
    assert chat_model.model == "claude-2"
    assert chat_model.max_tokens == 1024
    assert chat_model.temperature == 2


@pytest.mark.openai
def test_openai_chat_model_completion():
    model = OpenaiChatModel(
        model="gpt-4o",
        max_tokens=5,
        temperature=0.5,
    )
    response = model.complete(messages=[UserMessage("Hello!")])
    assert isinstance(response, AssistantMessage)
    assert isinstance(response.content, str)
    # TODO: test for num_tokens here


def test_chat_model_context(monkeypatch):
    monkeypatch.setenv("OPENAI_API_KEY", "test")
    chat_model = OpenaiChatModel("gpt-4")
    with chat_model:
        assert get_chat_model() is chat_model


def test_chat_model_context_within_context(monkeypatch):
    monkeypatch.setenv("OPENAI_API_KEY", "test")
    with OpenaiChatModel("gpt-4"):
        assert get_chat_model().model == "gpt-4"  # type: ignore[attr-defined]

        with OpenaiChatModel("gpt-5"):
            assert get_chat_model().model == "gpt-5"  # type: ignore[attr-defined]

        assert get_chat_model().model == "gpt-4"  # type: ignore[attr-defined]



import types
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Callable, Iterable, Iterator
from contextvars import ContextVar
from itertools import chain
from typing import Any, TypeVar, cast, get_origin, overload

from pydantic import ValidationError

from magentic._streamed_response import AsyncStreamedResponse, StreamedResponse
from magentic.chat_model.message import AssistantMessage, Message
from magentic.function_call import (
    AsyncParallelFunctionCall,
    FunctionCall,
    ParallelFunctionCall,
)
from magentic.streaming import AsyncStreamedStr, StreamedStr, achain, async_iter

R = TypeVar("R")

_chat_model_context: ContextVar["ChatModel | None"] = ContextVar(
    "chat_model", default=None
)


# TODO: Export all exceptions from `magentic.exceptions`
# TODO: Parent class with `output_message` attribute ?
class StringNotAllowedError(Exception):
    """Raised when a string is returned by the LLM but not allowed."""

    _MESSAGE = (
        "A string was returned by the LLM but is not an allowed output type."
        " Consider updating the allowed output types or modifying the prompt."
        " Model output: {model_output!r}"
    )

    def __init__(self, model_output: str):
        super().__init__(self._MESSAGE.format(model_output=model_output))
        self.output_message = AssistantMessage(model_output)


class FunctionCallNotAllowedError(Exception):
    """Raised when a FunctionCall is returned by the LLM but not allowed."""

    _MESSAGE = (
        "A function call was returned by the LLM but is not an allowed output type."
        " Consider updating the allowed output types or modifying the prompt."
        " FunctionCall: {function_call!r}"
    )

    def __init__(self, function_call: FunctionCall[Any]):
        super().__init__(self._MESSAGE.format(function_call=function_call))
        self.output_message = AssistantMessage(function_call)


class ObjectNotAllowedError(Exception):
    """Raised when a Python object is returned by the LLM but not allowed."""

    _MESSAGE = (
        "An object was returned by the LLM but is not an allowed output type."
        " Consider updating the allowed output types or modifying the prompt."
        " Object: {obj!r}"
    )

    def __init__(self, obj: Any):
        super().__init__(self._MESSAGE.format(obj=obj))
        self.output_message = AssistantMessage(obj)


class UnknownToolError(Exception):
    """Raised when the LLM returns a tool call for an unknown tool."""

    _MESSAGE = (
        "The LLM returned a tool call for a tool name that is not recognized."
        " Tool name: {tool_name!r}"
    )

    def __init__(self, output_message: Message[Any], tool_call_id: str, tool_name: str):
        super().__init__(self._MESSAGE.format(tool_name=tool_name))
        self.output_message = output_message
        self.tool_call_id = tool_call_id


# TODO: Move this to same file where it is raised
class ToolSchemaParseError(Exception):
    """Raised when the LLM output could not be parsed by the tool schema."""

    _MESSAGE = (
        "Failed to parse the LLM output into the tool schema."
        " Consider making the output type more lenient or enabling retries."
        " Model output: {model_output!r}"
    )

    def __init__(
        self,
        output_message: Message[Any],
        tool_call_id: str,
        validation_error: ValidationError,
    ):
        super().__init__(self._MESSAGE.format(model_output=output_message.content))
        self.output_message = output_message
        self.tool_call_id = tool_call_id
        self.validation_error = validation_error


# TODO: Move this into _parsing
# TODO: Make this a stream class with a close method and context management
def parse_stream(stream: Iterator[Any], output_types: Iterable[type[R]]) -> R:
    """Parse and validate the LLM output stream against the allowed output types."""
    output_type_origins = [get_origin(type_) or type_ for type_ in output_types]
    # TODO: option to error/warn/ignore extra objects
    # TODO: warn for degenerate output types ?
    obj = next(stream)
    if isinstance(obj, StreamedStr):
        if StreamedResponse in output_type_origins:
            return cast(R, StreamedResponse(chain([obj], stream)))
        if StreamedStr in output_type_origins:
            return cast(R, obj)
        if str in output_type_origins:
            return cast(R, str(obj))
        raise StringNotAllowedError(obj.truncate(100))
    if isinstance(obj, FunctionCall):
        if StreamedResponse in output_type_origins:
            return cast(R, StreamedResponse(chain([obj], stream)))
        if ParallelFunctionCall in output_type_origins:
            return cast(R, ParallelFunctionCall(chain([obj], stream)))
        if FunctionCall in output_type_origins:
            # TODO: Check that FunctionCall type matches ?
            return cast(R, obj)
        raise FunctionCallNotAllowedError(obj)
    if isinstance(obj, tuple(output_type_origins)):
        return cast(R, obj)
    raise ObjectNotAllowedError(obj)


async def aparse_stream(
    stream: AsyncIterator[Any], output_types: Iterable[type[R]]
) -> R:
    """Async version of `parse_stream`."""
    output_type_origins = [get_origin(type_) or type_ for type_ in output_types]
    obj = await anext(stream)
    if isinstance(obj, AsyncStreamedStr):
        if AsyncStreamedResponse in output_type_origins:
            return cast(R, AsyncStreamedResponse(achain(async_iter([obj]), stream)))
        if AsyncStreamedStr in output_type_origins:
            return cast(R, obj)
        if str in output_type_origins:
            return cast(R, await obj.to_string())
        raise StringNotAllowedError(await obj.truncate(100))
    if isinstance(obj, FunctionCall):
        if AsyncStreamedResponse in output_type_origins:
            return cast(R, AsyncStreamedResponse(achain(async_iter([obj]), stream)))
        if AsyncParallelFunctionCall in output_type_origins:
            return cast(R, AsyncParallelFunctionCall(achain(async_iter([obj]), stream)))
        if FunctionCall in output_type_origins:
            return cast(R, obj)
        raise FunctionCallNotAllowedError(obj)
    if isinstance(obj, tuple(output_type_origins)):
        return cast(R, obj)
    raise ObjectNotAllowedError(obj)


class ChatModel(ABC):
    """An LLM chat model."""

    @overload
    @abstractmethod
    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: None = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[str]: ...

    @overload
    @abstractmethod
    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: Iterable[type[R]] = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[R]: ...

    @abstractmethod
    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        # TODO: Set default of R to str in Python 3.13
        output_types: Iterable[type[R | str]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> AssistantMessage[str] | AssistantMessage[R]:
        """Request an LLM message."""
        ...

    @overload
    @abstractmethod
    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: None = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[str]: ...

    @overload
    @abstractmethod
    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: Iterable[type[R]] = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[R]: ...

    @abstractmethod
    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[R | str]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> AssistantMessage[str] | AssistantMessage[R]:
        """Async version of `complete`."""
        ...

    def __enter__(self) -> None:
        self.__token = _chat_model_context.set(self)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        _chat_model_context.reset(self.__token)



import collections.abc
import json
import typing
from collections import OrderedDict
from typing import Annotated, Any, Generic, TypeVar, get_origin

import pytest
from pydantic import BaseModel, Field, create_model

from magentic._pydantic import ConfigDict, with_config
from magentic.chat_model.function_schema import (
    AnyFunctionSchema,
    AsyncIterableFunctionSchema,
    BaseModelFunctionSchema,
    DictFunctionSchema,
    FunctionCallFunctionSchema,
    IterableFunctionSchema,
)
from magentic.function_call import FunctionCall
from magentic.streaming import async_iter
from magentic.typing import is_origin_subclass


@pytest.mark.parametrize(
    ("type_", "json_schema"),
    [
        (
            str,
            {
                "name": "return_str",
                "parameters": {
                    "properties": {"value": {"title": "Value", "type": "string"}},
                    "required": ["value"],
                    "type": "object",
                },
            },
        ),
        (
            int,
            {
                "name": "return_int",
                "parameters": {
                    "properties": {"value": {"title": "Value", "type": "integer"}},
                    "required": ["value"],
                    "type": "object",
                },
            },
        ),
        (
            bool | str,
            {
                "name": "return_bool_or_str",
                "parameters": {
                    "properties": {
                        "value": {
                            "title": "Value",
                            "anyOf": [{"type": "boolean"}, {"type": "string"}],
                        }
                    },
                    "required": ["value"],
                    "type": "object",
                },
            },
        ),
        (
            list[str],
            {
                "name": "return_list_of_str",
                "parameters": {
                    "properties": {
                        "value": {
                            "title": "Value",
                            "type": "array",
                            "items": {"type": "string"},
                        }
                    },
                    "required": ["value"],
                    "type": "object",
                },
            },
        ),
        (
            list[str | int],
            {
                "name": "return_list_of_str_or_int",
                "parameters": {
                    "properties": {
                        "value": {
                            "title": "Value",
                            "type": "array",
                            "items": {
                                "anyOf": [{"type": "string"}, {"type": "integer"}]
                            },
                        }
                    },
                    "required": ["value"],
                    "type": "object",
                },
            },
        ),
        (
            str | None,
            {
                "name": "return_str_or_null",
                "parameters": {
                    "properties": {
                        "value": {
                            "title": "Value",
                            "anyOf": [{"type": "string"}, {"type": "null"}],
                        }
                    },
                    "required": ["value"],
                    "type": "object",
                },
            },
        ),
        (
            dict[str, int],
            {
                "name": "return_dict_of_str_to_int",
                "parameters": {
                    "properties": {
                        "value": {
                            "title": "Value",
                            "type": "object",
                            "additionalProperties": {"type": "integer"},
                        }
                    },
                    "required": ["value"],
                    "type": "object",
                },
            },
        ),
    ],
)
def test_any_function_schema(type_, json_schema):
    function_schema = AnyFunctionSchema(type_)
    assert function_schema.dict() == json_schema


any_function_schema_args_test_cases = [
    (str, '{"value": "Dublin"}', "Dublin"),
    (int, '{"value": 42}', 42),
    (bool | str, '{"value": true}', True),
    (bool | str, '{"value": "Dublin"}', "Dublin"),
    (list[str], '{"value": ["Dublin", "London"]}', ["Dublin", "London"]),
    (list[str | int], '{"value": ["Dublin", 42]}', ["Dublin", 42]),
    (str | None, '{"value": "Dublin"}', "Dublin"),
    (str | None, '{"value": null}', None),
]


@pytest.mark.parametrize(
    ("type_", "args_str", "expected_args"), any_function_schema_args_test_cases
)
def test_any_function_schema_parse_args(type_, args_str, expected_args):
    parsed_args = AnyFunctionSchema(type_).parse_args(args_str)
    assert parsed_args == expected_args


@pytest.mark.parametrize(
    ("type_", "args_str", "expected_args"), any_function_schema_args_test_cases
)
async def test_any_function_schema_aparse_args(type_, args_str, expected_args):
    parsed_args = await AnyFunctionSchema(type_).aparse_args(async_iter(args_str))
    assert parsed_args == expected_args


@pytest.mark.parametrize(
    ("type_", "expected_args_str", "args"), any_function_schema_args_test_cases
)
def test_any_function_schema_serialize_args(type_, expected_args_str, args):
    serialized_args = AnyFunctionSchema(type_).serialize_args(args)
    assert json.loads(serialized_args) == json.loads(expected_args_str)


@pytest.mark.parametrize(
    ("type_", "json_schema"),
    [
        (
            list[str],
            {
                "name": "return_list_of_str",
                "parameters": {
                    "properties": {
                        "value": {
                            "title": "Value",
                            "type": "array",
                            "items": {"type": "string"},
                        }
                    },
                    "required": ["value"],
                    "type": "object",
                },
            },
        ),
        (
            list,
            {
                "name": "return_list",
                "parameters": {
                    "properties": {
                        "value": {
                            "title": "Value",
                            "type": "array",
                            "items": {},
                        }
                    },
                    "required": ["value"],
                    "type": "object",
                },
            },
        ),
        (
            typing.Iterable[str],
            {
                "name": "return_iterable_of_str",
                "parameters": {
                    "properties": {
                        "value": {
                            "title": "Value",
                            "type": "array",
                            "items": {"type": "string"},
                        }
                    },
                    "required": ["value"],
                    "type": "object",
                },
            },
        ),
        (
            collections.abc.Iterable[str],
            {
                "name": "return_iterable_of_str",
                "parameters": {
                    "properties": {
                        "value": {
                            "title": "Value",
                            "type": "array",
                            "items": {"type": "string"},
                        }
                    },
                    "required": ["value"],
                    "type": "object",
                },
            },
        ),
    ],
)
def test_iterable_function_schema(type_, json_schema):
    function_schema = IterableFunctionSchema(type_)
    assert function_schema.dict() == json_schema


iterable_function_schema_args_test_cases = [
    (list[str], '{"value": ["One", "Two"]}', ["One", "Two"]),
    (typing.Iterable[str], '{"value": ["One", "Two"]}', ["One", "Two"]),
    (collections.abc.Iterable[str], '{"value": ["One", "Two"]}', ["One", "Two"]),
]


@pytest.mark.parametrize(
    ("type_", "args_str", "expected_args"), iterable_function_schema_args_test_cases
)
def test_iterable_function_schema_parse_args(type_, args_str, expected_args):
    parsed_args = IterableFunctionSchema(type_).parse_args(args_str)
    assert isinstance(parsed_args, get_origin(type_))
    assert list(parsed_args) == expected_args


@pytest.mark.parametrize(
    ("type_", "expected_args_str", "args"), iterable_function_schema_args_test_cases
)
def test_iterable_function_schema_serialize_args(type_, expected_args_str, args):
    serialized_args = IterableFunctionSchema(type_).serialize_args(args)
    assert json.loads(serialized_args) == json.loads(expected_args_str)


@pytest.mark.parametrize(
    ("type_", "json_schema"),
    [
        (
            typing.AsyncIterable[str],
            {
                "name": "return_asynciterable_of_str",
                "parameters": {
                    "properties": {
                        "value": {
                            "title": "Value",
                            "type": "array",
                            "items": {"type": "string"},
                        }
                    },
                    "required": ["value"],
                    "type": "object",
                },
            },
        ),
        (
            typing.AsyncIterable,
            {
                "name": "return_asynciterable",
                "parameters": {
                    "properties": {
                        "value": {
                            "title": "Value",
                            "type": "array",
                            "items": {},
                        }
                    },
                    "required": ["value"],
                    "type": "object",
                },
            },
        ),
        (
            collections.abc.AsyncIterable[str],
            {
                "name": "return_asynciterable_of_str",
                "parameters": {
                    "properties": {
                        "value": {
                            "title": "Value",
                            "type": "array",
                            "items": {"type": "string"},
                        }
                    },
                    "required": ["value"],
                    "type": "object",
                },
            },
        ),
    ],
)
def test_async_iterable_function_schema(type_, json_schema):
    function_schema = AsyncIterableFunctionSchema(type_)
    assert function_schema.dict() == json_schema


async_iterable_function_schema_args_test_cases = [
    (typing.AsyncIterable[str], '{"value": ["One", "Two"]}', ["One", "Two"]),
    (collections.abc.AsyncIterable[str], '{"value": ["One", "Two"]}', ["One", "Two"]),
]


@pytest.mark.parametrize(
    ("type_", "args_str", "expected_args"),
    async_iterable_function_schema_args_test_cases,
)
async def test_async_iterable_function_schema_aparse_args(
    type_, args_str, expected_args
):
    parsed_args = await AsyncIterableFunctionSchema(type_).aparse_args(
        async_iter(args_str)
    )
    assert isinstance(parsed_args, get_origin(type_))
    assert [arg async for arg in parsed_args] == expected_args


@pytest.mark.parametrize(
    ("type_", "expected_args_str", "args"),
    async_iterable_function_schema_args_test_cases,
)
async def test_async_iterable_function_schema_aserialize_args(
    type_, expected_args_str, args
):
    serialized_args = await AsyncIterableFunctionSchema(type_).aserialize_args(
        async_iter(args)
    )
    assert json.loads(serialized_args) == json.loads(expected_args_str)


class User(BaseModel):
    name: str
    age: int


@pytest.mark.parametrize(
    ("type_", "json_schema"),
    [
        (
            dict,
            {
                "name": "return_dict",
                "parameters": {
                    "properties": {},
                    "type": "object",
                },
            },
        ),
        (
            dict[str, Any],
            {
                "name": "return_dict_of_str_to_any",
                "parameters": {
                    "properties": {},
                    "type": "object",
                },
            },
        ),
        (
            dict[str, int],
            {
                "name": "return_dict_of_str_to_int",
                "parameters": {
                    "additionalProperties": {"type": "integer"},
                    "properties": {},
                    "type": "object",
                },
            },
        ),
        (
            dict[str, User],
            {
                "name": "return_dict_of_str_to_user",
                "parameters": {
                    "$defs": {
                        "User": {
                            "properties": {
                                "name": {"title": "Name", "type": "string"},
                                "age": {"title": "Age", "type": "integer"},
                            },
                            "required": ["name", "age"],
                            "title": "User",
                            "type": "object",
                        }
                    },
                    "additionalProperties": {"$ref": "#/$defs/User"},
                    "properties": {},
                    "type": "object",
                },
            },
        ),
    ],
)
def test_dict_function_schema(type_, json_schema):
    function_schema = DictFunctionSchema(type_)
    assert function_schema.dict() == json_schema


dict_function_schema_args_test_cases = [
    (dict, '{"name": "Alice"}', {"name": "Alice"}),
    (dict[str, Any], '{"name": "Alice"}', {"name": "Alice"}),
    (dict[str, str], '{"name": "Alice"}', {"name": "Alice"}),
    (dict[str, int], '{"age": 99}', {"age": 99}),
    (
        dict[str, User],
        '{"alice": {"name": "Alice", "age": 99}}',
        {"alice": User(name="Alice", age=99)},
    ),
    (OrderedDict[str, int], '{"age": 99}', OrderedDict({"age": 99})),
]


@pytest.mark.parametrize(
    ("type_", "args_str", "expected_args"), dict_function_schema_args_test_cases
)
def test_dict_function_schema_parse_args(type_, args_str, expected_args):
    parsed_args = DictFunctionSchema(type_).parse_args(args_str)
    assert parsed_args == expected_args
    assert is_origin_subclass(type_, type(parsed_args))


@pytest.mark.parametrize(
    ("type_", "expected_args_str", "args"), dict_function_schema_args_test_cases
)
def test_dict_function_schema_serialize_args(type_, expected_args_str, args):
    serialized_args = DictFunctionSchema(type_).serialize_args(args)
    assert json.loads(serialized_args) == json.loads(expected_args_str)


@pytest.mark.parametrize(
    ("type_", "json_schema"),
    [
        (
            create_model(
                "User",
                name=(str, ...),
                age=(int, Field(description="Age", gt=0, examples=[10, 20])),
            ),
            {
                "name": "return_user",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "name": {"title": "Name", "type": "string"},
                        "age": {
                            "description": "Age",
                            "examples": [10, 20],
                            "exclusiveMinimum": 0,
                            "title": "Age",
                            "type": "integer",
                        },
                    },
                    "required": ["name", "age"],
                },
            },
        ),
        (
            create_model(
                "User",
                __config__=ConfigDict(openai_strict=True),
                name=(str, ...),
                age=(int, Field(description="Age", gt=0, examples=[10, 20])),
            ),
            {
                "name": "return_user",
                "parameters": {
                    "additionalProperties": False,
                    "properties": {
                        "age": {
                            "description": "Age",
                            "examples": [10, 20],
                            "exclusiveMinimum": 0,
                            "title": "Age",
                            "type": "integer",
                        },
                        "name": {"title": "Name", "type": "string"},
                    },
                    "required": ["name", "age"],
                    "title": "User",
                    "type": "object",
                },
                "strict": True,
            },
        ),
    ],
)
def test_base_model_function_schema(type_, json_schema):
    function_schema = BaseModelFunctionSchema(type_)
    assert function_schema.dict() == json_schema


def test_base_model_function_schema_parse_args():
    function_schema = BaseModelFunctionSchema(User)
    assert function_schema.parse_args('{"name": "Alice", "age": 99}') == User(
        name="Alice", age=99
    )


def test_base_model_function_schema_generic_model():
    T = TypeVar("T")

    class User(BaseModel, Generic[T]):
        name: str
        age: T

    function_schema = BaseModelFunctionSchema(User[int])

    assert function_schema.name == "return_user_int"
    assert function_schema.dict() == {
        "name": "return_user_int",
        "parameters": {
            "type": "object",
            "properties": {
                # TODO: Remove "title" keys from schema
                "name": {"title": "Name", "type": "string"},
                "age": {"title": "Age", "type": "integer"},
            },
            "required": ["name", "age"],
        },
    }
    assert function_schema.parse_args('{"name": "Alice", "age": 99}') == User(
        name="Alice", age=99
    )


def plus(a: int, b: int) -> int:
    return a + b


def plus_no_type_hints(a, b):
    return a + b


def plus_default_value(a: int, b: int = 3) -> int:
    return a + b


def plus_with_args(a: int, *args: int) -> int:
    return a + sum(args)


def plus_with_args_no_type_hints(a, *args):
    return a + sum(args)


def plus_with_kwargs(a: int, **kwargs: int) -> int:
    return a + sum(kwargs.values())


def plus_with_kwargs_no_type_hints(a, **kwargs):
    return a + sum(kwargs.values())


def plus_with_annotated(
    a: Annotated[int, Field(description="First number")],
    b: Annotated[int, Field(description="Second number")],
) -> int:
    return a + b


def return_constant() -> int:
    return 100


class IntModel(BaseModel):
    value: int


def plus_with_basemodel(a: IntModel, b: IntModel) -> IntModel:
    return IntModel(value=a.value + b.value)


@with_config(ConfigDict(openai_strict=True))
def plus_with_config_openai_strict(a: int, b: int) -> int:
    return a + b


@pytest.mark.parametrize(
    ("function", "json_schema"),
    [
        (
            plus,
            {
                "name": "plus",
                "parameters": {
                    "type": "object",
                    "properties": {
                        # TODO: Remove "title" keys from schema
                        "a": {"title": "A", "type": "integer"},
                        "b": {"title": "B", "type": "integer"},
                    },
                    "required": ["a", "b"],
                },
            },
        ),
        (
            plus_no_type_hints,
            {
                "name": "plus_no_type_hints",
                "parameters": {
                    "type": "object",
                    "properties": {
                        # TODO: Remove "title" keys from schema
                        "a": {"title": "A"},
                        "b": {"title": "B"},
                    },
                    "required": ["a", "b"],
                },
            },
        ),
        (
            plus_default_value,
            {
                "name": "plus_default_value",
                "parameters": {
                    "type": "object",
                    "properties": {
                        # TODO: Remove "title" keys from schema
                        "a": {
                            "title": "A",
                            "type": "integer",
                        },
                        "b": {
                            "default": 3,
                            "title": "B",
                            "type": "integer",
                        },
                    },
                    "required": ["a"],
                },
            },
        ),
        (
            plus_with_args,
            {
                "name": "plus_with_args",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "a": {"title": "A", "type": "integer"},
                        "args": {
                            "default": [],
                            "items": {"type": "integer"},
                            "title": "Args",
                            "type": "array",
                        },
                    },
                    "required": ["a"],
                },
            },
        ),
        (
            plus_with_args_no_type_hints,
            {
                "name": "plus_with_args_no_type_hints",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "a": {"title": "A"},
                        "args": {
                            "default": [],
                            "items": {},
                            "title": "Args",
                            "type": "array",
                        },
                    },
                    "required": ["a"],
                },
            },
        ),
        (
            plus_with_kwargs,
            {
                "name": "plus_with_kwargs",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "a": {"title": "A", "type": "integer"},
                        "kwargs": {
                            "additionalProperties": {"type": "integer"},
                            "default": {},
                            "title": "Kwargs",
                            "type": "object",
                        },
                    },
                    "required": ["a"],
                },
            },
        ),
        (
            plus_with_kwargs_no_type_hints,
            {
                "name": "plus_with_kwargs_no_type_hints",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "a": {"title": "A"},
                        "kwargs": {
                            "default": {},
                            "title": "Kwargs",
                            "type": "object",
                        },
                    },
                    "required": ["a"],
                },
            },
        ),
        (
            plus_with_annotated,
            {
                "name": "plus_with_annotated",
                "parameters": {
                    "type": "object",
                    "properties": {
                        # TODO: Remove "title" keys from schema
                        "a": {
                            "title": "A",
                            "type": "integer",
                            "description": "First number",
                        },
                        "b": {
                            "title": "B",
                            "type": "integer",
                            "description": "Second number",
                        },
                    },
                    "required": ["a", "b"],
                },
            },
        ),
        (
            plus_with_basemodel,
            {
                "name": "plus_with_basemodel",
                "parameters": {
                    "$defs": {
                        "IntModel": {
                            "properties": {
                                "value": {"title": "Value", "type": "integer"}
                            },
                            "required": ["value"],
                            "title": "IntModel",
                            "type": "object",
                        }
                    },
                    "properties": {
                        "a": {"$ref": "#/$defs/IntModel"},
                        "b": {"$ref": "#/$defs/IntModel"},
                    },
                    "required": ["a", "b"],
                    "type": "object",
                },
            },
        ),
        (
            plus_with_config_openai_strict,
            {
                "name": "plus_with_config_openai_strict",
                "parameters": {
                    "additionalProperties": False,
                    "properties": {
                        "a": {"title": "A", "type": "integer"},
                        "b": {"title": "B", "type": "integer"},
                    },
                    "required": ["a", "b"],
                    "title": "FuncModel",
                    "type": "object",
                },
                "strict": True,
            },
        ),
        (
            return_constant,
            {
                "name": "return_constant",
                "parameters": {"type": "object", "properties": {}},
            },
        ),
    ],
)
def test_function_call_function_schema(function, json_schema):
    function_schema = FunctionCallFunctionSchema(function)
    assert function_schema.dict() == json_schema


def test_function_call_function_schema_with_default_value():
    function_schema = FunctionCallFunctionSchema(plus_default_value)
    output = function_schema.parse_args('{"a": 1}')
    assert isinstance(output, FunctionCall)
    assert output() == 4


function_call_function_schema_args_test_cases = [
    (plus, '{"a": 1, "b": 2}', FunctionCall(plus, 1, 2)),
    (
        plus_no_type_hints,
        '{"a": 1, "b": 2}',
        FunctionCall(plus_no_type_hints, 1, 2),
    ),
    (plus_default_value, '{"a": 1}', FunctionCall(plus_default_value, 1)),
    (plus_with_args, '{"a": 1, "args": [2, 3]}', FunctionCall(plus_with_args, 1, 2, 3)),
    (
        plus_with_args_no_type_hints,
        '{"a": 1, "args": [2, 3]}',
        FunctionCall(plus_with_args_no_type_hints, 1, 2, 3),
    ),
    (
        plus_with_kwargs,
        '{"a": 1, "kwargs": {"b": 2, "c": 3}}',
        FunctionCall(plus_with_kwargs, 1, b=2, c=3),
    ),
    (
        plus_with_kwargs_no_type_hints,
        '{"a": 1, "kwargs": {"b": 2, "c": 3}}',
        FunctionCall(plus_with_kwargs_no_type_hints, 1, b=2, c=3),
    ),
    (
        plus_with_annotated,
        '{"a": 1, "b": 2}',
        FunctionCall(plus_with_annotated, 1, 2),
    ),
    (
        plus_with_basemodel,
        '{"a": {"value": 1}, "b": {"value": 2}}',
        FunctionCall(plus_with_basemodel, IntModel(value=1), IntModel(value=2)),
    ),
    (
        return_constant,
        "{}",
        FunctionCall(return_constant),
    ),
]


@pytest.mark.parametrize(
    ("function", "args_str", "expected_args"),
    [
        *function_call_function_schema_args_test_cases,
        # Anthropic message stream returns empty string for function call with no arguments
        (return_constant, "", FunctionCall(return_constant)),
    ],
)
def test_function_call_function_schema_parse_args(function, args_str, expected_args):
    parsed_args = FunctionCallFunctionSchema(function).parse_args(args_str)
    assert parsed_args == expected_args


@pytest.mark.parametrize(
    ("function", "expected_args_str", "args"),
    function_call_function_schema_args_test_cases,
)
def test_function_call_function_schema_serialize_args(
    function, expected_args_str, args
):
    serialized_args = FunctionCallFunctionSchema(function).serialize_args(args)
    assert json.loads(serialized_args) == json.loads(expected_args_str)


@pytest.mark.parametrize(
    ("function", "expected_args_str", "args"),
    [
        (
            plus,
            '{"a": "non-int", "b": {"value": 5}}',
            FunctionCall(plus, "non-int", IntModel(value=5)),  # type: ignore[arg-type]
        ),
    ],
)
def test_function_call_function_schema_serialize_invalid_args(
    function, expected_args_str, args
):
    """Invalid function arguments should serialize so LLM errors can be resubmitted."""
    serialized_args = FunctionCallFunctionSchema(function).serialize_args(args)
    assert json.loads(serialized_args) == json.loads(expected_args_str)
